from __future__ import absolute_import
from botocore.exceptions import ClientError
import codecs, os, sys, json
from collections import OrderedDict

import trollius as asyncio
from trollius import From, Return

from secret.vault import Kms

BOTO_DEFAULT = ''
ENCODING = 'utf-8'
PY3 = (sys.version_info.major == 3)

class Storage(object):
    def list(self, **kw): pass
    def get(self, key, **kw): pass
    def put(self, key, value, **kw): pass

class S3(Storage):
    """
    A single bucket holds multiple projects.
    """
    def __init__(self, session, vault, vaultkey, region, prefix, project, env='default'):
        self.client = session.client('s3')
        self.bucket = vault
        self.vault = Kms(session=session, key=vaultkey)
        self.vaultkey = vaultkey
        self.region = region
        self.prefix = prefix
        self.project = project
        self.env = env + '/'

        if self.prefix:
            self.prefix += '/'

    @asyncio.coroutine
    def create_bucket(self, name=None, **kw):
        # administrator creates bucket for projects
        name = name or self.bucket
        result = self.client.create_bucket(Bucket=name)
        raise Return(result)

    @asyncio.coroutine
    def setup(self, vault, project, **kw):
        # project-user creates a folder inside bucket
        if not all([vault, project]):
            sys.exit("Error! Vault and/or project undefined")
        project_s3 = project.rstrip('/')+'/'
        try:
            result = yield From(self.get(project_s3, in_group_check=False))
            res = "Error! A project with this name already exists"
        except ClientError as e:
            result = yield From(self.put(project_s3, '', prefixify=False))
            self.project.save(dict(vault=vault, project=project, key=self.vaultkey, region=self.region))
            res = "Success! Project created. Configuration stored in %s"%self.project.name
        raise Return(res)

    @asyncio.coroutine
    def list_backend(self, prefix=None, **kw):
        prefix = prefix or self.prefixify('')
        raise Return(self.client.list_objects(Bucket=self.bucket, Prefix=prefix, MaxKeys=1000, Delimiter=kw.get('delimiter', BOTO_DEFAULT)))

    @asyncio.coroutine
    def list(self, key=None, **kw):
        key = self.prefixify(key)
        response = yield From(self.list_backend(prefix=key, **kw))
        raise Return([self.prefixify(k['Key'], reverse=True) for k in response.get('Contents', [])])

    @asyncio.coroutine
    def config(self, key=None, **kw):
        contents = yield From(self.list_backend(prefix=self.prefixify(key)))
        contents = contents.get('Contents', [])
        kw.setdefault('in_group_check', True)
        tasks = [self.get(key=obj['Key'], **kw) for obj in contents]
        try:
            results = yield From(asyncio.gather(*tasks))
        except ValueError as e:
            raise Return({})
        keys = [k['Key'].split('/')[-1] for k in contents]
        raise Return(OrderedDict(zip(keys, results)))

    @asyncio.coroutine
    def put_backend(self, **kw):
        raise Return(self.client.put_object(**kw))

    @asyncio.coroutine
    def put(self, key, value, **kw):
        if value is None:
            raise Return("Error! No value provided.")

        if kw.get('prefixify', True):
            key = self.prefixify(key)
        is_file = False
        is_binary = False # TODO: --binary
        mode = 'rb'
        if os.path.isfile(value):
            value = codecs.open(os.path.expandvars(os.path.expanduser(value)), mode=mode, encoding=ENCODING).read().rstrip('\n')
            is_file = True
        value = value.encode(ENCODING)
        data = self.vault.encrypt(value, is_file=is_file, is_binary=is_binary)
        data['name'] = key
        result = yield From(self.put_backend(Bucket=self.bucket, Key=key, Body=json.dumps(data)))
        if result.get('ResponseMetadata').get('HTTPStatusCode') == 200:
            raise Return("Success! Wrote: %s"%key)
        else:
            raise Return(result)

    @asyncio.coroutine
    def get_backend(self, **kw):
        raise Return(self.client.get_object(**kw))

    @asyncio.coroutine
    def get(self, key, **kw):
        if key is None:
            raise Return("Error! No key provided.")

        key = self.prefixify(key)
        extra = {}
        if kw.get('version'):
            extra['VersionId'] = kw.get('version')
        try:
            result = yield From(self.get_backend(Bucket=self.bucket, Key=key, **extra))
        except Exception as e:
            if kw.get('in_group_check', True):
                # check for grouped keys
                key = key.rstrip('/')+'/'
                kw['in_group_check'] = True
                result = yield From(self.config(key=key, **kw))
                if not result:
                    result = "Error! The specified key does not exist."
                raise Return(result)
            raise
        body = json.loads(result['Body'].read().decode(ENCODING))
        data = self.vault.decrypt(body)
        is_binary = body.get('is_binary', False)
        if not is_binary:
            data = data.decode(ENCODING)
        raise Return(data)

    @asyncio.coroutine
    def delete_backend(self, **kw):
        raise Return(self.client.delete_object(**kw))

    @asyncio.coroutine
    def delete(self, key, **kw):
        key = self.prefixify(key)
        result = yield From(self.delete_backend(Bucket=self.bucket, Key=key))
        if result.get('ResponseMetadata').get('HTTPStatusCode') == 204:
            raise Return("Success! Deleted: %s"%key)
        else:
            raise Return(result)

    @asyncio.coroutine
    def versions(self, **kw):
        prefix = kw.get('prefix', self.prefix)
        response = self.client.list_object_versions(Bucket=self.bucket, Prefix=self.prefix)
        versions = []
        for k in response.get('Versions', []):
            key = self.prefixify(k['Key'], reverse=True)
            if kw.get('key') and kw.get('key')!=key: continue
            versions.append(dict(key=key, version=k['VersionId'], is_latest=k['IsLatest'], modified=k['LastModified']))
        raise Return(versions)

    @asyncio.coroutine
    def envs(self, **kw):
        contents = yield From(self.list_backend(prefix=self.prefix, delimiter='/'))
        raise Return(list(k['Prefix'].split('/')[1] for k in contents.get('CommonPrefixes', [])))

    def prefixify(self, key, reverse=False):
        if reverse:
            if self.prefix and self.env:
                key = key.replace('{}{}'.format(self.prefix, self.env), '')
        elif key is not None:
            if self.env:
                if self.prefix not in key and self.env not in key:
                    key = '{}{}{}'.format(self.prefix, self.env, key)
            else:
                if self.prefix not in key:
                    key = '{}{}'.format(self.prefix, key)
        key = key or ''
        key = key.replace('//','/')
        return key



import os, sys
import json

class Project:
    def __init__(self, name):
        self.name = name
        self.data = None

    def load(self):
        if self.data is None:
            try:
                data = json.loads(open(self.name, 'r').read())
            except IOError as e:
                data = {}
            self.data = data
        return self.data

    def save(self, c):
        settings = self.load()
        settings.update(**c)
        with open(self.name, 'w') as f:
            f.write(json.dumps(settings))
        self.data = None

def get_project(name):
    return Project(name=name)



from __future__ import absolute_import
from base64 import b64encode, b64decode
from secret.algo import AesEax

ENCODING = 'utf-8'

def tob64(value):
    return b64encode(value).decode(ENCODING)

class Vault:
    def encrypt(self, value, **kw): pass
    def decrypt(self, value, **kw): pass

class KmsDataKey:
    def __init__(self, vault, key, ciphertext=None, num_bytes=64):
        if ciphertext:
            res = vault.client.decrypt(CiphertextBlob=ciphertext)
            self.wrapped_key = ciphertext
        else:
            res = vault.client.generate_data_key(KeyId=key, NumberOfBytes=num_bytes)
            self.wrapped_key = res['CiphertextBlob']
        self.data_key = res['Plaintext'][:32]
        self.hmac_key = res['Plaintext'][32:]

class Kms(Vault):
    def __init__(self, session, key="alias/secret"):
        self.client = session.client('kms')
        self.key = key
        self.cipher = AesEax

    def keyclass(self, **kw):
        return KmsDataKey(vault=self, key=self.key, **kw)

    def encrypt(self, value, **kw):
        datakey = self.keyclass()

        cipher = self.cipher(datakey.data_key)
        c_text, nonce, tag = cipher.encrypt(value)

        hmac = cipher.hmac(datakey.hmac_key, c_text)

        data = {}
        data['key'] = tob64(datakey.wrapped_key)
        data['contents'] = tob64(c_text)
        data['nonce'] = tob64(nonce)
        data['tag'] = tob64(tag)
        data['hmac'] = hmac
        data['is_file'] = kw.get('is_file', False)
        data['is_binary'] = kw.get('is_binary', False)
        return data

    def decrypt(self, data, **kw):
        datakey = self.keyclass(ciphertext=b64decode(data['key']))

        cipher = self.cipher(datakey.data_key)
        assert (cipher.hmac(datakey.hmac_key, b64decode(data['contents'])) == data['hmac'])

        return cipher.decrypt(b64decode(data['contents']),
                nonce=b64decode(data['nonce']),
                tag=b64decode(data['tag']),)



from setuptools import setup, find_packages, Command

import os, sys, subprocess

class TestCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        if os.getenv('PYTHON_VERSION') == 'secret-python3':# .travis.yml
            subprocess.call(['bash conv_py2_to_py3.bash',], shell=True)

        raise SystemExit(
            subprocess.call(['py.test',]))

install_requires=['boto3>=0.0.18', 'pycryptodome>=3.1','tabulate>=0.7.5','pytest>=2.8.5']
if sys.version_info.major == 2:
    install_requires.append('trollius>=1.0.4')

setup(
    name="secret",
    version='0.6.1',
    description='Secret is for storing secrets. Backed by Amazon Web Services: IAM for access policies, KMS for encryption keys and S3 for storage',
    keywords = 'secret secrets project vault aws amazon cloud',
    author = 'Jussi Vaihia',
    author_email = 'jussi.vaihia@futurice.com',
    license='Apache2',
    url='https://github.com/futurice/secret',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
    ],
    install_requires=install_requires,
    packages = ["secret"],
    include_package_data = True,
    cmdclass = {
        'test': TestCommand,
    },
    entry_points={
        'console_scripts': [
            'secret = secret.secret:runner'
            ]
        }
)
